#!/usr/bin/env -S npx tsx

// src/bootstrap/config.ts
import path2 from "node:path";
import { cwd } from "node:process";
import defu from "defu";

// src/log.ts
import consola, { LogLevels } from "consola";
function isLogger(m) {
  return typeof m === "object" && m !== null && "log" in m;
}
function getConsolaLevel(level) {
  switch (level) {
    case "error":
      return 0;
    case "warn":
      return 1;
    case "normal":
      return 2;
    case "info":
      return 3;
    case "debug":
      return 4;
    case "trace":
      return 5;
    default:
      return 3;
  }
}
function getLogger(name) {
  let config2;
  try {
    config2 = getConfig();
  } catch (e) {
  }
  consola.level = getConsolaLevel(
    config2?.logger?.level ?? defaultConfig.logger.level
  );
  if (process.env.DEBUG) {
    consola.level = LogLevels.debug;
  }
  return name ? consola.withTag(name) : consola;
}

// src/bootstrap/file-module.ts
import fs2 from "node:fs/promises";
import { readdir } from "node:fs/promises";
import path from "node:path";
import { pathToFileURL } from "node:url";

// src/plainstack-fs.ts
import fs from "node:fs/promises";
async function fileExists(filePath) {
  try {
    await fs.access(filePath);
    return true;
  } catch {
    return false;
  }
}
async function directoryExists(dirPath) {
  try {
    await fs.access(dirPath);
    return true;
  } catch {
    return false;
  }
}

// src/bootstrap/file-module.ts
async function loadModule(filePath, load2) {
  const log = getLogger("load-module");
  if (!await fileExists(filePath)) {
    log.debug(`tried loading module, ${filePath} does not exist`);
    return void 0;
  }
  const module = await import(pathToFileURL(filePath).toString());
  const result = {
    namedExports: {}
  };
  if ("default" in module) {
    const defaultExport = module.default;
    if (defaultExport && typeof defaultExport === "object" && "default" in defaultExport) {
      result.defaultExport = await load2(defaultExport.default);
    } else {
      result.defaultExport = await load2(defaultExport);
    }
  }
  for (const [key, value] of Object.entries(module)) {
    if (key !== "default") {
      result.namedExports[key] = await load2(value);
    }
  }
  if (result.defaultExport)
    log.debug(`successfully loaded default export in module ${filePath}`);
  if (Object.keys(result.namedExports).length)
    log.debug(`successfully loaded named exports in module ${filePath}`);
  return result;
}
async function loadModulesFromDir(baseDir, load2, extensions = [".ts", ".tsx"], currentDir = baseDir) {
  const modules = [];
  const log = getLogger("manifest");
  if (!await directoryExists(currentDir)) {
    log.debug(`tried loading modules, directory ${currentDir} does not exist`);
    return [];
  }
  const files = await readdir(currentDir);
  log.debug(`found ${files.length} files in ${currentDir}`);
  for (const file of files) {
    const absolutePath = path.join(currentDir, file);
    const stat = await fs2.stat(absolutePath);
    if (stat.isDirectory()) {
      log.debug(`found directory: ${absolutePath}`);
      const subModules = await loadModulesFromDir(
        baseDir,
        load2,
        extensions,
        absolutePath
      );
      modules.push(...subModules);
    } else if (stat.isFile()) {
      log.debug(`found file: ${absolutePath}`);
      if (!extensions.includes(path.extname(file))) {
        log.debug(`skipping file ${absolutePath} with unsupported extension`);
        continue;
      }
      const relativePath = path.relative(baseDir, absolutePath);
      const result = await loadModule(absolutePath, load2);
      if (!result) continue;
      modules.push({
        defaultExport: result.defaultExport,
        namedExports: result.namedExports,
        filename: path.parse(file).name,
        extension: path.extname(file),
        absolutePath,
        relativePath
      });
    }
  }
  return modules;
}

// src/bootstrap/config.ts
var config;
function isConfig(m) {
  return typeof m === "object" && m !== null && "nodeEnv" in m;
}
async function load(m) {
  if (!isConfig(m)) throw new Error("invalid config module");
  return m;
}
var defaultConfig = {
  nodeEnv: "production",
  logger: {
    level: "normal"
  },
  http: {
    port: 3e3,
    staticPath: "/public"
  },
  paths: {
    routes: "app/routes",
    commands: "app/commands",
    jobs: "app/jobs",
    schedules: "app/schedules",
    schema: "app/config/schema.ts",
    styles: "assets/styles.css",
    assets: "assets",
    out: ".out",
    seed: "database/seed.ts",
    migrations: "database/migrations",
    database: "app/config/database.ts",
    http: "app/config/http.ts",
    queue: "app/config/queue.ts",
    mailer: "app/config/mailer.ts",
    logger: "app/config/logger.ts"
  }
};
function withDefaultConfig(config2) {
  return defu(config2, defaultConfig);
}
async function loadConfig() {
  const log = getLogger("load-config");
  const fileModule = await loadModule(
    path2.join(cwd(), "plainstack.config.ts"),
    load
  );
  if (!fileModule) throw new Error("no plainstack.config.ts found");
  if (!fileModule.defaultExport)
    throw new Error("no default export found in plainstack.config.ts");
  config = withDefaultConfig(fileModule.defaultExport);
  log.debug("config loaded from file", config);
}
function getConfig() {
  if (!config)
    throw new Error("config not loaded, make sure to call loadConfig() first");
  return config;
}
async function loadAndGetConfig() {
  const log = getLogger("load-config");
  if (config) {
    log.debug("return memoized config");
    return config;
  }
  await loadConfig();
  return getConfig();
}

// src/bootstrap/get.ts
var appLoader;
function setAppLoader(loader) {
  appLoader = loader;
}
async function get(keys) {
  const log = getLogger("app-loader");
  if (!appLoader)
    throw new Error(
      "appLoader not set, make sure to call registerAppLoader() first"
    );
  log.debug(`getting app components for ${keys.join(", ")}`);
  const results = await Promise.all(
    keys.map(async (key) => {
      try {
        const loader = appLoader?.[key];
        if (!loader) {
          log.warn(`no loader found for ${key}`);
          return [key, void 0];
        }
        log.debug(`loading app component: ${key}`);
        const result = await loader();
        log.debug(`successfully loaded app component: ${key}`);
        return [key, result];
      } catch (error) {
        log.error(`failed to load app component: ${key}`, error);
        return [key, void 0];
      }
    })
  );
  return Object.fromEntries(results);
}
async function getOrThrow(keys) {
  const log = getLogger("app-loader");
  if (!appLoader) {
    throw new Error("appLoader not set. call setAppLoader first.");
  }
  log.debug(`getting app components (or throw) for ${keys.join(", ")}`);
  const result = await get(keys);
  const missingKeys = keys.filter((key) => result[key] === void 0);
  if (missingKeys.length > 0) {
    throw new Error(`missing app components: ${missingKeys.join(", ")}`);
  }
  return result;
}

// src/bootstrap/load.ts
import { join } from "node:path";
import { consola as consola2 } from "consola";

// src/command/command.ts
import chokidar from "chokidar";
import { $ as $2 } from "execa";
import fs3 from "fs-extra";

// src/database/database.ts
import { Migrator, sql } from "kysely";
import { TSFileMigrationProvider } from "kysely-ctl";
function isDatabase(db) {
  return typeof db === "object" && db !== null && "schema" in db && "selectFrom" in db;
}

// src/database/seed.ts
function isSeeder(m) {
  return typeof m === "object" && m !== null && "run" in m;
}

// src/web/http.ts
import expressListEndpoints from "express-list-endpoints";
import expressRateLimit from "express-rate-limit";
import morgan from "morgan";

// src/id.ts
import { createId } from "@paralleldrive/cuid2";

// src/web/file-router.ts
import express from "express";

// src/web/plain-response.ts
import "@kitajs/html";
import { renderToStream } from "@kitajs/html/suspense";

// src/web/handler.ts
function isHandler(m) {
  return typeof m === "function";
}

// src/web/file-router.ts
function getFileRoutesFromFileModules(modules) {
  const log = getLogger("file-router");
  const fileRoutes = [];
  for (const module of modules) {
    const fileRoute = {
      filePath: module.relativePath,
      GET: void 0,
      POST: void 0
    };
    if (!module.namedExports?.GET && !module.namedExports?.POST && !module.defaultExport) {
      throw new Error(
        `invalid file route module at ${module.absolutePath}, no GET or POST export found`
      );
    }
    if (module.namedExports?.GET) {
      log.debug(`found GET export in ${module.absolutePath}`);
      fileRoute.GET = module.namedExports.GET;
    }
    if (module.namedExports?.POST) {
      log.debug(`found POST export in ${module.absolutePath}`);
      fileRoute.POST = module.namedExports.POST;
    }
    if (module.defaultExport && !fileRoute.GET) {
      log.debug(
        `found default export in ${module.absolutePath}, assuming GET export`
      );
      fileRoute.GET = module.defaultExport;
    }
    fileRoutes.push(fileRoute);
  }
  return fileRoutes;
}

// src/web/http.ts
function isHttp(m) {
  return typeof m === "function";
}

// src/command/info.ts
import { $ } from "execa";

// src/command/command.ts
function isCommand(cmd) {
  return typeof cmd === "object" && cmd !== null && "run" in cmd;
}

// src/env.ts
import dotenv from "dotenv";
import z from "zod";
function test() {
  return process.env.NODE_ENV === "test";
}

// src/job.ts
import {
  defineQueue as definePlainjobsQueue,
  defineWorker
} from "plainjobs";
function isJob(job) {
  return typeof job === "object" && job !== null && "run" in job && "name" in job && typeof job.run === "function";
}
function isSchedule(schedule) {
  return typeof schedule === "object" && schedule !== null && "run" in schedule && "name" in schedule && typeof schedule.run === "function";
}
function isQueue(q) {
  return typeof q === "object" && q !== null && "add" in q && "schedule" in q;
}
async function work(queue, jobs, schedules) {
  const log = getLogger("work");
  if (!Object.values(jobs).length && !Object.values(schedules).length) {
    log.warn("started worker, but no jobs or schedules to run");
  }
  for (const schedule of Object.values(schedules)) {
    log.debug(`scheduling ${schedule.name} to run at ${schedule.cron}`);
    queue.schedule(schedule.name, { cron: schedule.cron });
  }
  log.debug(`starting worker for jobs: ${Object.keys(jobs).join(", ")}`);
  if (Object.values(schedules).length) {
    log.debug(
      `starting worker for schedules: ${Object.keys(schedules).join(", ")}`
    );
  }
  const workables = [...Object.values(jobs), ...Object.values(schedules)];
  const workers = workables.map(
    (w) => defineWorker(w.name, w.run, { queue, logger: log })
  );
  await Promise.all(workers.map((worker) => worker.start()));
}

// src/mail.ts
import nodemailer from "nodemailer";
var devMailer = nodemailer.createTransport({
  jsonTransport: true
});
function isMailer(m) {
  return typeof m === "object" && m !== null && "sendMail" in m;
}

// src/bootstrap/load.ts
function getDatabaseLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.database);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isDatabase(m))
        throw new Error(
          `default export in ${modulePath} is not a valid database`
        );
      return m;
    });
    if (!fileModule)
      throw new Error(`no default export found in ${modulePath}`);
    if (!fileModule.defaultExport)
      throw new Error(`no default export found in ${modulePath}`);
    return fileModule.defaultExport;
  };
}
function getServerLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.http);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isHttp(m))
        throw new Error(
          `default export in ${modulePath} is not a valid http server`
        );
      return m;
    });
    if (!fileModule)
      throw new Error(`no default export found in ${modulePath}`);
    if (!fileModule.defaultExport)
      throw new Error(`no default export found in ${modulePath}`);
    return {
      start: async () => {
        const app = await fileModule.defaultExport?.(config2);
        return app;
      }
    };
  };
}
function getQueueLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.queue);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isQueue(m))
        throw new Error(`default export in ${modulePath} is not a valid queue`);
      return m;
    });
    if (!fileModule)
      throw new Error(`no default export found in ${modulePath}`);
    if (!fileModule.defaultExport)
      throw new Error(`no default export found in ${modulePath}`);
    const namedExports = {};
    for (const [key, value] of Object.entries(fileModule.namedExports)) {
      if (!isQueue(value))
        throw new Error(
          `named export ${key} in ${modulePath} is not a valid queue`
        );
      namedExports[key] = value;
    }
    return {
      default: fileModule.defaultExport,
      named: namedExports
    };
  };
}
function getMailerLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.mailer);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isMailer(m))
        throw new Error(
          `default export in ${modulePath} is not a valid mailer`
        );
      return m;
    });
    if (!fileModule)
      throw new Error(`no default export found in ${modulePath}`);
    if (!fileModule.defaultExport)
      throw new Error(`no default export found in ${modulePath}`);
    const namedExports = {};
    for (const [key, value] of Object.entries(fileModule.namedExports)) {
      if (!isMailer(value))
        throw new Error(
          `named export ${key} in ${modulePath} is not a valid mailer`
        );
      namedExports[key] = test() ? devMailer : value;
    }
    if (test()) {
      const log = getLogger("app-loader");
      log.warn("NODE_ENV=test, skipping mailer");
    }
    return {
      default: test() ? devMailer : fileModule.defaultExport,
      named: namedExports
    };
  };
}
function getJobsLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.jobs);
    const fileModules = await loadModulesFromDir(
      modulePath,
      async (m) => {
        if (!isJob(m))
          throw new Error(`default export in ${modulePath} is not a valid job`);
        return m;
      },
      [".ts"]
    );
    const jobs = {};
    for (const fileModule of fileModules) {
      if (!fileModule.defaultExport)
        throw new Error(`no default export found in ${modulePath}`);
      jobs[fileModule.filename] = fileModule.defaultExport;
    }
    return jobs;
  };
}
function getCommandsLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.commands);
    const fileModules = await loadModulesFromDir(
      modulePath,
      async (m) => {
        if (!isCommand(m))
          throw new Error(
            `default export in ${modulePath} is not a valid command`
          );
        return m;
      },
      [".ts"]
    );
    const commands = {};
    for (const fileModule of fileModules) {
      if (!fileModule.defaultExport)
        throw new Error(`no default export found in ${modulePath}`);
      commands[fileModule.filename] = fileModule.defaultExport;
    }
    return commands;
  };
}
function getSchedulesLoader(opts) {
  return async () => {
    const log = getLogger("app-loader");
    if (test()) {
      log.warn("NODE_ENV=test, skipping schedules");
      return {};
    }
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.schedules);
    const fileModules = await loadModulesFromDir(
      modulePath,
      async (m) => {
        if (!isSchedule(m))
          throw new Error(
            `default export in ${modulePath} is not a valid schedule`
          );
        return m;
      },
      [".ts"]
    );
    const schedules = {};
    for (const fileModule of fileModules) {
      if (!fileModule.defaultExport)
        throw new Error(`no default export found in ${modulePath}`);
      schedules[fileModule.filename] = fileModule.defaultExport;
    }
    return schedules;
  };
}
function getSeederLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.seed);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isSeeder(m))
        throw new Error(
          `default export in ${modulePath} is not a valid seeder`
        );
      return m;
    });
    if (!fileModule)
      throw new Error(`no default export found in ${modulePath}`);
    if (!fileModule.defaultExport)
      throw new Error(`no default export found in ${modulePath}`);
    return fileModule.defaultExport;
  };
}
function getRoutesLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.routes);
    const fileModules = await loadModulesFromDir(
      modulePath,
      async (m) => {
        if (!isHandler(m))
          throw new Error(
            `default export in ${modulePath} is not a valid handler`
          );
        return m;
      },
      [".tsx"]
    );
    return getFileRoutesFromFileModules(fileModules);
  };
}
function getLoggersLoader(opts) {
  return async () => {
    const { cwd: cwd2, config: config2 } = opts ?? {
      cwd: process.cwd(),
      config: await loadAndGetConfig()
    };
    const modulePath = join(cwd2, config2.paths.logger);
    const fileModule = await loadModule(modulePath, async (m) => {
      if (!isLogger(m)) throw new Error(`${modulePath} is not a valid logger`);
      return m;
    });
    if (!fileModule) throw new Error(`no module found in ${modulePath}`);
    const loggers = [];
    for (const [key, value] of Object.entries(fileModule.namedExports)) {
      if (!isLogger(value))
        throw new Error(
          `named export ${key} in ${modulePath} is not a valid logger`
        );
      loggers.push(value);
    }
    if (fileModule.defaultExport) loggers.push(fileModule.defaultExport);
    for (const logger of loggers) {
      consola2.addReporter(logger);
    }
    return loggers;
  };
}
function memoize(loader) {
  const log = getLogger("load-memoize");
  let result;
  return async () => {
    if (result) {
      log.debug("returning memoized app component", { loader: loader.name });
      return result;
    }
    log.debug("loading app component", { loader: loader.name });
    result = await loader();
    return result;
  };
}
function registerAppLoader(opts) {
  const log = getLogger("app-loader");
  log.debug("registering app loader", { opts });
  const database = memoize(getDatabaseLoader(opts));
  const server = memoize(getServerLoader(opts));
  const queue = memoize(getQueueLoader(opts));
  const mailer = memoize(getMailerLoader(opts));
  const jobs = memoize(getJobsLoader(opts));
  const commands = memoize(getCommandsLoader(opts));
  const schedules = memoize(getSchedulesLoader(opts));
  const seeder = memoize(getSeederLoader(opts));
  const routes = memoize(getRoutesLoader(opts));
  const loggers = memoize(getLoggersLoader(opts));
  setAppLoader({
    database,
    server,
    queue,
    mailer,
    jobs,
    commands,
    schedules,
    seeder,
    routes,
    loggers
  });
}

// src/bin/bin-work.ts
registerAppLoader();
async function main() {
  const log = getLogger("work");
  const config2 = await loadAndGetConfig();
  const { queue, jobs, schedules } = await getOrThrow([
    "queue",
    "jobs",
    "schedules"
  ]);
  if (!queue.default)
    throw new Error(
      `can not start job worker without a queue, make sure you have a export default defineQueue() in ${config2.paths.queue}`
    );
  await work(queue.default, jobs, schedules);
  log.info("\u26A1\uFE0F worker started");
}
void main();
//# sourceMappingURL=bin-work.js.map